#!/usr/bin/env python3
import argparse
import sys
import os
import textwrap
from mge_masker.mge_masker_functions import find_mges, create_mge_gff_file, mask_mges, merge_mges, print_default_regex_patterns

def is_valid_file(parser, arg):
    if not os.path.isfile(arg):
        parser.error('The file {} does not exist!'.format(arg))
    else:
        # File exists so return the filename
        return arg


def parse_args():
    description = textwrap.dedent(
        """
        A module to find MGEs in a rich sequence file and mask regions corresponding to the MGEs in a pseudogenome alignment.
        
        The find_mges command searches a gbk or embl file for features that have MGE-associated annotations.
        It writes a GFF file containing the positions of the matched features.
        
        The mask_mges command takes a GFF file produced using the find_mges command and masks those regions in all sequences of a pseudogenome alignment based on the reference sequence used to find MGEs.
        """
    )

    parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawDescriptionHelpFormatter)
    subparsers = parser.add_subparsers(
        help='The following commands are available. Type mge_masker <COMMAND> -h for more help on a specific commands',
        dest='command'
    )

    subparsers.required = True

    find_mges = subparsers.add_parser('find_mges',
        help='Search a rich sequence file for features annotated with text that suggests a MGE-associated element'
    )

    find_mges.add_argument('-g', '--genome_file_path', type=lambda x: is_valid_file(parser, x), help='path to a genome file', required=True)
    find_mges.add_argument('-f', '--file_format',
                        help='genome file format',
                        default = 'genbank',
                        choices=['genbank', 'embl']
                        )
    find_mges.add_argument('-i', '--merge_interval', help='The maximum distance between MGEs when performing the merging step (Default 1000bp)', default=1000, type = int)
    find_mges.add_argument('-m', '--mge_file_path', type=lambda x: is_valid_file(parser, x), help='path to a file containing regex MGE annotations')
    
    mask_mges = subparsers.add_parser('mask_mges',
        help='Mask regions from a pseudogenome alignment with the regions in a GFF file produced using the find_mges command'
    )
    mask_mges.add_argument('-f', '--fasta_path', help='path to either a single fasta reference or a pseudogenome alignment file', type=lambda x: is_valid_file(parser, x), required=True)
    mask_mges.add_argument('-g', '--gff_file_path', help='path to a gff file containing MGE regions to be masked', type=lambda x: is_valid_file(parser, x), required=True)
    mask_mges.add_argument('-m', '--masking_character', help='character used to mask (default: N)', default='N')
    
    default_matches = subparsers.add_parser('default_matches',
        help='Show the default regex patterns used when searching for MGEs'
    )
    
    return(parser.parse_args())





def main():
    options = parse_args()
    if (options.command == 'find_mges'):
        mges = find_mges(options.genome_file_path, options.file_format, options.mge_file_path)
        print(f'Found {len(mges)} MGEs')
        merged_mges = merge_mges(mges, options.merge_interval)
        if len(merged_mges) < len(mges):
            print(f'{len(mges)} MGEs were merged into {len(merged_mges)}')
        else:
            print('No MGEs were merged')
        gff_file_path = create_mge_gff_file(options.genome_file_path, options.file_format, merged_mges)
        print(f'GFF file written to {gff_file_path}')
    elif (options.command == 'mask_mges'):
        masked_alignment_file_path = mask_mges(options.fasta_path, options.gff_file_path, options.masking_character)
        print(f'Masked alignment file written to {masked_alignment_file_path}')
    elif(options.command == 'default_matches'):
        print_default_regex_patterns()
    
 

if __name__ == "__main__":
    main()




from Bio import SeqIO
try:
  from Bio.Alphabet import generic_dna
except ImportError: # Alphabet gone
  generic_dna is None

from Bio.Seq import Seq 
from Bio.SeqRecord import SeqRecord 
import inspect
import copy
import sys
import os
import re
from pathlib import Path
import csv
from collections import OrderedDict


def get_mge_patterns(mge_file_path = None):
  # read text file with pattern match regular expressions and return a list of compiled regexs
  if not mge_file_path:
    mge_file_path = os.path.join(os.path.dirname(__file__), "mge_patterns.txt")
  mge_patterns = []
  with open(mge_file_path) as mge_file:
    for line in mge_file.readlines():
      pattern = re.compile(line.rstrip())
      mge_patterns.append(pattern)
  return(mge_patterns)

def get_features(genome_file_path, file_format):
  record = SeqIO.read(genome_file_path, file_format)
  return(record.features)

def create_gff_line(accession, mge):
  description = mge['description']
  description_string = ';'.join([f'{match}="{",".join(description[match])}"' for match in description])
  return(f'{accession}\t.\t{mge["type"]}\t{mge["start"]}\t{mge["end"]}\t.\t{mge["strand"]}\t.\t{description_string}\n')

def search_features_for_patterns(features, mge_patterns):
  mges = []
  for feature in features:
    qualifiers = OrderedDict()
    for qualifier in feature.qualifiers:
      if qualifier in ['product','note']:
          qualifiers[qualifier] = feature.qualifiers[qualifier]
    if len(qualifiers) > 0:
      matches = OrderedDict()
      for qualifier in qualifiers:
        for qualifier_value in qualifiers[qualifier]:
          for pattern in mge_patterns:
            if pattern.match(qualifier_value):
              if qualifier not in matches:
                matches[qualifier] = []
              matches[qualifier].append(qualifier_value)
      if len(matches) > 0:
        if feature.strand:
          if feature.strand == 1:
            strand = "+"
          elif feature.strand == -1:
            strand = "-"
        else:
          strand = "."
        mges.append(
          {
            "type": feature.type,
            "start": feature.location.start.position + 1,
            "end": feature.location.end.position,
            "strand": strand,
            "description": matches
          }
        )
  return(mges)

def find_mges(genome_file_path, file_format, mge_file_path = None):
  """
  returns mges as a list of dicts with the keys: type, start, end, strand, description
  """
  features = get_features(genome_file_path, file_format)
  mge_patterns = get_mge_patterns(mge_file_path)
  mges = search_features_for_patterns(features, mge_patterns)
  return(mges)


def merge_mges(mges, merge_interval):
  previous_mge = {}
  merged_mge = { "type": None, "start": None, "end": None, "strand": None, "description": {}}
  number_mges_merged = 1
  merged_mges = []
  for mge in mges:
    if not previous_mge:
      previous_mge = mge
    else:
      # check if this mge should be merged with previous
      if mge['start'] - previous_mge['end'] <= merge_interval: # merge if less than interval
        if not merged_mge['start']:
          merged_mge['start'] = previous_mge['start']
        merged_mge['end'] = mge['end']
        if not merged_mge['strand']:
          merged_mge['strand'] = previous_mge['strand']
        if mge['strand'] != merged_mge['strand']:
            merged_mge['strand'] = '.'

        if not merged_mge['description']:
          merged_mge['description'] = copy.deepcopy(previous_mge['description'])
        # merge the descriptions
        for match in mge['description']:
          if match not in merged_mge['description']:
            merged_mge['description'][match] = []
          merged_mge['description'][match].extend(mge['description'][match])
        number_mges_merged += 1
      else:
        # check if a mge has been merged
        if merged_mge['start']:
          merged_mge['type'] = f"{number_mges_merged} merged features"
          merged_mges.append(merged_mge)
          # reset merged_mge
          merged_mge = { "type": None, "start": None, "end": None, "strand": None, "description": {}}
          number_mges_merged = 1
        else:
          merged_mges.append(previous_mge)
      previous_mge = mge
    
    # Final merge
  if merged_mge['start']:
    merged_mge['type'] = f"{number_mges_merged} merged features"
    merged_mges.append(merged_mge)
    # reset merged_mge
    merged_mge = { "type": None, "start": None, "end": None, "strand": None, "description": {}}
  else:
    merged_mges.append(previous_mge)

  return(merged_mges)
        

def create_mge_gff_file(genome_file_path, file_format, mges):
  record = SeqIO.read(genome_file_path, file_format)
  gff_file_path = Path(genome_file_path).with_suffix('.gff')
  with open(gff_file_path, "w") as gff_file:
    for mge in mges:
      gff_file.write(create_gff_line(record.name, mge))
  return(gff_file_path)

def extract_position_ranges_from_gff_file(gff_file_path):
  with open(gff_file_path) as gff_file:
    reader = csv.reader(gff_file, delimiter='\t')
    start_and_end_positions = []
    for row in reader:
      try:
        start_and_end_positions.append((int(row[3]), int(row[4])))
      except IndexError as error:
        sys.exit(f"{error}\n\nThe GFF file is not formatted correctly\nProblem with the line below. Please check tabs and the format\n{' '.join(row)}")
  return(start_and_end_positions)

def find_min_and_max_positions(positions):
  min_pos = min([pos[0] for pos in positions])
  max_pos= max([pos[1] for pos in positions])
  return(min_pos, max_pos)

def check_alignment(alignment_length, gff_ranges):
  # get max position specified in the GFF file
  min_pos, max_pos = find_min_and_max_positions(gff_ranges)
  # check max pos doesn't exceed alignment length
  if alignment_length < max_pos:
    sys.exit(f'The maximum position {max_pos} specified in the GFF file exceeds the alignment length')

def find_fasta_length(fasta_path):
  with open(fasta_path) as input_alignment:
    current_sequence = ""
    for line in input_alignment:
      if line.startswith(">"):
        if current_sequence != "":
          return(len(current_sequence)) # Multiple fasta sequence
        else: # first sequence
          next
      else:
        current_sequence += line.strip()
    return(len(current_sequence)) # Single fasta sequence

def mask_sequence(sequence, gff_ranges, masking_character):
  for gff_range in gff_ranges:
    start = gff_range[0] -1
    end = gff_range[1]
    difference = end - start
    sequence = f'{sequence[:start]}{masking_character*difference}{sequence  [end:]}'
  return(sequence)

def mask_mges(fasta_path, gff_file_path, masking_character):
  # find length of pseudoalignment
  alignment_length = find_fasta_length(fasta_path)  
  gff_ranges = extract_position_ranges_from_gff_file(gff_file_path)
  check_alignment(alignment_length, gff_ranges)

  masked_fasta_path = Path(fasta_path).with_suffix('.masked.fas') 
  with open(fasta_path) as input_alignment, open(masked_fasta_path, 'w') as output_alignment:
    sequence = ""
    index = 0
    for line in input_alignment:
      if line.startswith(">"):
        index += 1
        if index+1 > 0 and (index+1) % 10 == 0:
          print(f'Masking sequence {index+1}')
        if sequence != "": # previous sequence finished
          sequence = mask_sequence(sequence, gff_ranges, masking_character)
          if generic_dna: # deals with deprecate Alphabet
            record = SeqRecord(Seq(sequence, generic_dna), id=id, description = "")
          else:
            record = SeqRecord(Seq(sequence), id=id, description = "")

          output_alignment.write(record.format("fasta"))
          sequence = ""
        id = line.rstrip().replace(">", "")
      else:
        sequence += line.rstrip()
    sequence = mask_sequence(sequence, gff_ranges, masking_character)
    if generic_dna: # deals with deprecate Alphabet
      record = SeqRecord(Seq(sequence, generic_dna), id=id, description = "")
    else:
      record = SeqRecord(Seq(sequence), id=id, description = "")
    output_alignment.write(record.format("fasta"))

  return(masked_fasta_path)
  

def print_default_regex_patterns():
  mge_file_path = os.path.join(os.path.dirname(__file__), "mge_patterns.txt")
  with open(mge_file_path) as mge_file:
    for line in mge_file.readlines():
      print(line.rstrip())


  


import pytest
import os
import inspect
import tempfile
from collections import OrderedDict
from mge_masker.mge_masker_functions import find_mges, merge_mges, create_mge_gff_file

@pytest.fixture(scope="module")
def test_genome_file_path():
  test_genome_file_path = os.path.realpath(
    os.path.abspath(
      os.path.join(
        os.path.split(inspect.getfile( inspect.currentframe() ))[0],
        "data",
        "test.gbk"
      )
    )
  )
  return(test_genome_file_path)

@pytest.fixture(scope="module")
def expected_gff_file_path():
  expected_gff_file_path = os.path.realpath(
    os.path.abspath(
      os.path.join(
        os.path.split(inspect.getfile( inspect.currentframe() ))[0],
        "data",
        "test.gff"
      )
    )
  )
  return(expected_gff_file_path)

@pytest.fixture(scope="module")
def expected_mges():
  return (
    [
      {'type': 'CDS', 'start': 201, 'end': 400, 'strand': '+', 'description': OrderedDict([('note', ['Possible transposase based on homology'])])},
      {'type': 'CDS', 'start': 401, 'end': 500, 'strand': '-', 'description': OrderedDict([('product', ['Phage protein'])])},
      {'type': 'CDS', 'start': 501, 'end': 600, 'strand': '+', 'description': OrderedDict([('product', ['IS101'])])},
      {'type': 'CDS', 'start': 701, 'end': 800, 'strand': '+', 'description': OrderedDict([('product', ['Integrase'])])}
    ]
  )

@pytest.fixture(scope="module")
def test_mges_for_merging():
  return (
    [
      {'type': 'CDS', 'start': 1000, 'end': 2000, 'strand': '+', 'description': OrderedDict([('note', ['Possible transposase based on homology'])])},
      {'type': 'CDS', 'start': 2500, 'end': 3500, 'strand': '-', 'description': OrderedDict([('product', ['Phage protein'])])},
      {'type': 'CDS', 'start': 5000, 'end': 6000, 'strand': '+', 'description': OrderedDict([('product', ['IS101'])])},
      {'type': 'CDS', 'start': 7000, 'end': 8000, 'strand': '+', 'description': OrderedDict([('product', ['Integrase'])])},
      {'type': 'CDS', 'start': 8500, 'end': 9000, 'strand': '+', 'description': OrderedDict([('product', ['Repeat'])])},
      {'type': 'CDS', 'start': 12000, 'end': 13000, 'strand': '-', 'description': OrderedDict([('product', ['Tandem repeat'])])}
    ]
  )

@pytest.fixture(scope="module") 
def expected_mges_after_merging_1000_interval():
  return (
    [
      {
        'type': '2 merged features',
        'start': 1000, 'end': 3500, 'strand': '.',
        'description': OrderedDict([('note', ['Possible transposase based on homology']), ('product', ['Phage protein'])])
      },
      {
        'type': '3 merged features',
        'start': 5000, 'end': 9000, 'strand': '+',
        'description': OrderedDict([('product', ['IS101', 'Integrase', 'Repeat'])])
      },
      {
        'type': 'CDS', 
        'start': 12000, 'end': 13000, 'strand': '-',
        'description': OrderedDict([('product', ['Tandem repeat'])])
      }
    ]
  )
@pytest.fixture(scope="module") 
def expected_mges_after_merging_500_interval():
  return (
    [
      {
        'type': '2 merged features',
        'start': 1000, 'end': 3500, 'strand': '.',
        'description': OrderedDict([('note', ['Possible transposase based on homology']), ('product', ['Phage protein'])])
      },
      {'type': 'CDS', 'start': 5000, 'end': 6000, 'strand': '+', 'description': OrderedDict([('product', ['IS101'])])},
      {
        'type': '2 merged features',
        'start': 7000, 'end': 9000, 'strand': '+',
        'description': OrderedDict([('product', ['Integrase', 'Repeat'])])
      },
      {
        'type': 'CDS', 
        'start': 12000, 'end': 13000, 'strand': '-',
        'description': OrderedDict([('product', ['Tandem repeat'])])
      }
    ]
  )
  
def test_find_mges(test_genome_file_path, expected_mges):
  mges = find_mges(test_genome_file_path, "genbank")
  
  for index, mge in enumerate(mges):
    assert mge == expected_mges[index]


def test_merge_mges_500_interval(test_mges_for_merging, expected_mges_after_merging_500_interval):
  merged_mges = merge_mges(test_mges_for_merging, 500)
  for index, mge in enumerate(merged_mges):
    assert mge == expected_mges_after_merging_500_interval[index]

def test_merge_mges_1000_interval(test_mges_for_merging, expected_mges_after_merging_1000_interval):
  merged_mges = merge_mges(test_mges_for_merging, 1000)
  for index, mge in enumerate(merged_mges):
    assert mge == expected_mges_after_merging_1000_interval[index]
    
def test_create_file(test_genome_file_path, expected_gff_file_path):
  with tempfile.NamedTemporaryFile(mode = "w") as temp_genome:
    with open(test_genome_file_path) as test_genome_file:
      genome_string = test_genome_file.read()
    temp_genome.write(genome_string)
    temp_genome.flush()
    mges = find_mges(temp_genome.name, "genbank")
    gff_file_path = create_mge_gff_file(temp_genome.name, "genbank", mges)
    with open(gff_file_path) as gff_file:
      gff_string = gff_file.read()
    with open(expected_gff_file_path) as expected_gff_file:
      expected_gff_string = expected_gff_file.read()
    assert gff_string == expected_gff_string
      




import pytest
import os
import inspect
import tempfile
from collections import OrderedDict
from mge_masker.mge_masker_functions import extract_position_ranges_from_gff_file, find_min_and_max_positions, mask_mges

@pytest.fixture(scope="module")
def test_gff_file_path():
  gff_file_path = os.path.realpath(
    os.path.abspath(
      os.path.join(
        os.path.split(inspect.getfile( inspect.currentframe() ))[0],
        "data",
        "test.gff"
      )
    )
  )
  return(gff_file_path)

@pytest.fixture(scope="module")
def test_alignment_file_path():
  alignment_file_path = os.path.realpath(
    os.path.abspath(
      os.path.join(
        os.path.split(inspect.getfile( inspect.currentframe() ))[0],
        "data",
        "test_alignment.fas"
      )
    )
  )
  return(alignment_file_path)

@pytest.fixture(scope="module")
def test_masked_alignment_file_path():
  masked_alignment_file_path = os.path.realpath(
    os.path.abspath(
      os.path.join(
        os.path.split(inspect.getfile( inspect.currentframe() ))[0],
        "data",
        "test_alignment.masked.fas"
      )
    )
  )
  return(masked_alignment_file_path)


def test_extract_boundaries_from_gff_file(test_gff_file_path):
  start_and_end_positions = extract_position_ranges_from_gff_file(test_gff_file_path)
  
  assert start_and_end_positions == [(201,400), (401,500), (501,600), (701,800)]


def test_extract_position_ranges_from_gff_file(test_gff_file_path):
  start_and_end_positions = extract_position_ranges_from_gff_file(test_gff_file_path)
  min, max = find_min_and_max_positions(start_and_end_positions)
  assert min == 201
  assert max == 800
  
def test_mask_mges(test_alignment_file_path, test_masked_alignment_file_path, test_gff_file_path):
  with tempfile.NamedTemporaryFile(mode = "w") as temp_alignment:
    with open(test_alignment_file_path) as test_alignment_file:
      alignment_string = test_alignment_file.read()
    temp_alignment.write(alignment_string)
    temp_alignment.flush()
    masked_pseudogenome_alignment_path= mask_mges(temp_alignment.name, test_gff_file_path, "N")
    with open(masked_pseudogenome_alignment_path) as masked_pseudogenome_alignment:
      masked_alignment_string = masked_pseudogenome_alignment.read()
    with open(test_masked_alignment_file_path) as test_masked_alignment_file:
      expected_masked_alignment_string = test_masked_alignment_file.read()
    assert masked_alignment_string == expected_masked_alignment_string
    




import setuptools
import os
import glob

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setuptools.setup(
    name="MGEmasker",
    version="0.1.9",
    author="Anthony Underwood",
    author_email="au3@sanger.ac.uk",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    license="GPLv3",
    packages=setuptools.find_packages(),
    package_data={'mge_masker': ['mge_patterns.txt', 'tests/data/*']},
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'mge_masker=mge_masker.run_mge_masker:main'
        ]
    },
    python_requires='>=3',
    install_requires=['biopython'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'coverage']
)